// ======================================================================
//
// Profile.cpp
// jeff grills
//
// copyright 1997 origin systems
// 
// ======================================================================

#include "First.h"

#if USE_PROFILE

#include "Profile.h"
#include "Profile.def"

#include <stdio.h>
#include <limits.h>

// ======================================================================

int          Profile::indxData;
int         *Profile::indx = &Profile::indxData;

Profile::ProfileEntry Profile::entry[MAX_ENTRIES];
Profile::ProfileCumm  Profile::cumm[MAX_CUMMULATIVE];

int          Profile::selectedLine;
const char  *Profile::base;

const char  *Profile::expand[MAX_EXPANDED];
int          Profile::expandCnt;

const char  *Profile::averageElem;
int          Profile::averageSum;
int          Profile::averageArray[MAX_AVERAGE];

int          Profile::hideCummulative;

char         Profile::buffer[BUFFER_SIZE];

const char * const Profile::startError = "Profile::start overflow";
const char * const Profile::stopError = "Profile::stop overflow";
const char * const Profile::contError = "Profile::cont overflow";

// ======================================================================

void Profile::reset(void)
{
	int i;

	// reset timing entry index
	indxData = 0;

	// clear out all the old cummulative data
	for (i = 0; i < MAX_CUMMULATIVE && cumm[i].text; i++)
	{
		cumm[i].calls = 0;
		cumm[i].time = 0;
	}
}

// ----------------------------------------------------------------------

int Profile::cummulativeHandle(const char *str)
{
	int i;

	for (i = 0; i < MAX_CUMMULATIVE && cumm[i].text && cumm[i].text != str; i++)
		;

	DEBUG_FATAL(i >= MAX_CUMMULATIVE, ("Profile::cummulativeHandle full"));
	cumm[i].text = str;

	return i;
}

// ----------------------------------------------------------------------

void Profile::dump(void)
{
	int i, lvl;
	
	for (i = 0, lvl = 0; i < indxData ; i++)
	{
		if (entry[i].op > 0)
			lvl++;
		else
			lvl--;
		
		DPRINT_LOG(("ProfileDump %-4d%-4d%s\n", lvl, entry[i].op, entry[i].text));
	}
}

// ----------------------------------------------------------------------

void Profile::update(void)
{
	char         buf[16];
	char         exp, here;
	char        *patch;
	const char  *selectedText; 
	int          expanded, expandable, outputLine;
	int          lvl, srch, slot;
	int          visStart, visStop;
	int          i, j;
	ProfileTime  dt, buildTime, frameTime;
	int          bufferOffset;
	int          pc;

	getProfileTime(&buildTime);

	// verify the recorded profile data for likely correctness
	DEBUG_FATAL(indxData > MAX_ENTRIES, ("Profile::update over max %d %d", indxData, MAX_ENTRIES));

	// must be an even number of entries
	if (indxData & 1)
	{
		dump();
		Fatal("Profile::update odd entries %d", indxData);
	}

	if (!PROFILE_ACTIVE || !indxData)
	{
		reset();
		return;
	}

	// verify starts and stops could match up
	for (i = 0, lvl = 0; i < indxData ; i++)
	{
		if (entry[i].op > 0)
			lvl++;
		else
			lvl--;
		
		if (lvl < 0)
		{
			dump();
			Fatal("Profile::update too many stops (%s)", entry[i].text);
		}
	}

	if (lvl > 0)
	{
		dump();
		Fatal("Profile::update missing stops");
	}

	// wipe the buffer clean
	bufferOffset = 0;
	buffer[0] = '\0';

	// set the range of visible elements to everything
	visStart = 0;
	visStop = indxData;

	// init selected text variables
	outputLine = 0;
	selectedText = NULL;

	// if we have a base for display, locate its range
	if (base)
	{
		for (i = 0; i < indxData && entry[i].text != base; i++)
			;

		if (i >= indxData)
			base = NULL;
		else
		{
			// mark the start of the visible range
			visStart = i;

			// find where this element ends
			for (srch = 1, j = i+1; j < indxData && srch > 0; j++)
				if (entry[j].op >= 0)
					srch++;
				else
					srch--;

			// mark the end of the visible range
			visStop = j-1;
		}
	}

	frameTime = entry[indxData-1].time - entry[0].time;
	if (frameTime == 0)
	{
		strcpy(buffer, "profile: 0 tick frame\n");
		reset();
		return;
	}

	// show non-cummulative timing stats
	for (i = 0, lvl = 0; i < indxData && lvl >= 0; i++)
	{
		// if this starts a normal time slice
		if (entry[i].op == 1)
		{
			// search for the current level to drop back to this level
			// also determine if this line is expandable by the user
			for (expandable = 0, srch = lvl+1, j = i+1; j < indxData && srch > lvl; j++)
				if (entry[j].op >= 0)
				{
					srch++;
					if (entry[j].op == 1)
						expandable = 1;
				}
				else
					srch--;

			// determine wether this block has been expanded by the user
			if (expandable)
				for (expanded=(lvl<1), srch = 0; !expanded && srch < expandCnt; srch++)
					expanded = (expand[srch] == entry[i].text);
			else
				expanded = 0;

			// determine what expanding character to show
			if (expandable)
				if (expanded)
					exp = '+';
				else
					exp = '-';
			else
				exp = ' ';

			// display if it is within the base range
			if (i >= visStart && i <= visStop)
			{
				dt = entry[j - 1].time - entry[i].time;
				pc = static_cast<int>(dt * 100 / frameTime);

				if (selectedLine == outputLine)
				{
					selectedText = entry[i].text;

					// mark the current line
					here = '>';

					// handle toggling of expansion for the current line
					if (expandable && PROFILE_KEY_EXPAND)
						if (expanded)
						{
							if (lvl >= 1)
							{
								// search for this expanded entry
								for (srch=0; srch < expandCnt && expand[srch] != entry[i].text; srch++)
									;

								// move the last element here to fill in hole
								if (srch < expandCnt)
									expand[srch] = expand[--expandCnt];
								expanded = 0;
							}
						}
						else
							if (expandCnt < MAX_EXPANDED)
							{
								expand[expandCnt++] = entry[i].text;
								expanded = 1;
							}
				}
				else
					here = ' ';

				// average timings for this line
				if (entry[i].text == averageElem)
				{
					// prime the array with the current value
					if (averageSum == 0)
						for (slot = 0; slot < MAX_AVERAGE; slot++)
						{
							averageArray[slot] = static_cast<int>(dt);
							averageSum += static_cast<int>(dt);
						}

					// recompute the new average sum
					slot = PROFILE_FRAME_COUNT & (MAX_AVERAGE-1);
					averageSum -= averageArray[slot];
					averageSum += (averageArray[slot] = static_cast<int>(dt));

					sprintf(buf, " %d=avg\n", averageSum / MAX_AVERAGE);
				}
				else
				{
					buf[0] = '\n';
					buf[1] = '\0';
				}

				if (bufferOffset < BUFFER_LIMIT)
				{
					outputLine++;
					bufferOffset += sprintf(buffer+bufferOffset, "%*c%-3d | %-8d%c%c %s%s",lvl*4,' ', pc, (int)dt, exp, here, entry[i].text, buf);
				}
			}

			// skip interior of non-expanded blocks
			if (expandable && !expanded)
			{
				i = j - 1;
				continue;
			}
		}

		// update the indention level
		if (entry[i].op > 0)
			lvl++;
		else
			lvl--;
	}

	if (PROFILE_KEY_AVERAGE)
	{
		averageSum = 0;
		if (averageElem == selectedText)
			averageElem = NULL;
		else
			averageElem = selectedText;
	}

	if (PROFILE_KEY_BASE)
		if (base == selectedText)
			base = NULL;
		else
			base = selectedText;

	if (PROFILE_KEY_CUMMULATIVE)
		hideCummulative ^= 1;

	if (PROFILE_KEY_UP && selectedLine)
		selectedLine--;

	if (PROFILE_KEY_DOWN)
		selectedLine++;

	if (selectedLine >= outputLine)
		selectedLine = outputLine - 1;

	if (!base && bufferOffset < BUFFER_LIMIT)
	{
		patch = buffer+bufferOffset+1;	
		bufferOffset += sprintf(buffer+bufferOffset, "            profile update(%d)\n", indxData);
	}
	else
		patch = NULL;

	if (!hideCummulative && bufferOffset < BUFFER_LIMIT)
	{
		int skipped;

		// print cummulative times for routines with multiple calls
		for (skipped = 0, i = 0; bufferOffset < BUFFER_LIMIT && i < MAX_CUMMULATIVE && cumm[i].text; i++)
			if (cumm[i].calls)
			{
				// skip a line between straight and cummulative timings
				if (!skipped)
				{
					buffer[bufferOffset++] = '\n';
					buffer[bufferOffset] = '\0';
					skipped = 1;
				}

				bufferOffset += sprintf(buffer+bufferOffset,"%-8d  calls: %-4d %s\n", (int)cumm[i].time, cumm[i].calls, cumm[i].text);
			}
	}

	// go back and patch in the total profile update time
	getProfileTime(&dt);
	dt -= buildTime;
	if (patch)
		patch[sprintf(patch, "%d", (int)dt)] = ' ';

	reset();
}

// ----------------------------------------------------------------------

void Profile::draw(void)
{
	PROFILE_OUTPUT(buffer);
}

// ======================================================================

#endif
